//  Init.c
// 
//  Version	Date	 Author	 Description
//	--------------------------------------------------------------------------
//  0.01	24/06/01 BWW	 Initial implementation

#include <c:\psx\include\libps.h>
#include "planner.h"
#include "funcprototypes.h"

extern	Agent_Struct			Agent;
extern	Key_Struct				Key;
extern	Start_Struct			Start;
extern	Exit_Struct				Exit;
extern	Door_Struct				Door[];
extern	Bridge_Struct			Bridge[];
extern	Route_Struct		 	Route;
extern	Preconditions_Struct	Preconditions;
extern	Effects_Struct			Effects;
extern	Plan_Struct				Plan;


// Standard Text
extern	char 	TurnMsg[];
extern	long	Turn;

extern	char 	Environment[10][20];

// Flag whether to print Debug information
extern	int		Debug;

// Sprite structure
extern	GsSPRITE 	MessageSprite;

// Ordering table variables
extern	GsOT		WorldOT[2];
extern	GsOT_TAG	OTTags[2][1<<OT_LENGTH];

// Define the packet work area
extern	PACKET		GpuPacketArea[2][PACKETMAX2];

// Which buffer variable
extern	int			activeBuff;

extern	int	EnvSpeed;
extern	int	EnvComplexity;
extern	int EnvStartLocation;

extern	int	BRIDGENUM;
extern	int	DOORNUM;



//	--------------------------------------------------------------------------
// 	Initialise the Preconditions and Effects 
//	--------------------------------------------------------------------------
//	Description:	Set up the preconditions and the effects that will be used
//					by the plannner to represent Goals and Actions
//
//	Parameters:		None
//	Returns:		None

void PreconditionEffectInit()
	{
	
	Preconditions.condition[0].condition =		PRECON_AT;
	Preconditions.condition[0].parameter1 = 	PARAM_OBJECTAGENT;
	Preconditions.condition[0].value1 = 		NULL;
	Preconditions.condition[0].parameter2 = 	PARAM_XPOS;
	Preconditions.condition[0].value2 = 		NULL;
	Preconditions.condition[0].parameter3 = 	PARAM_YPOS;
	Preconditions.condition[0].value3 = 		NULL;
	
	Preconditions.condition[1].condition =		PRECON_DOORUNLOCKED;
	Preconditions.condition[1].parameter1 = 	PARAM_OBJECTDOOR;
	Preconditions.condition[1].value1 = 		NULL;
	Preconditions.condition[1].parameter2 = 	PARAM_LOCKSTATUS;
	Preconditions.condition[1].value2 = 		NULL;
	Preconditions.condition[1].parameter3 = 	NULL;
	Preconditions.condition[1].value3 = 		NULL;

	Preconditions.condition[2].condition =		PRECON_DOOROPEN;
	Preconditions.condition[2].parameter1 = 	PARAM_OBJECTDOOR;
	Preconditions.condition[2].value1 = 		NULL;
	Preconditions.condition[2].parameter2 = 	PARAM_OPENSTATUS;
	Preconditions.condition[2].value2 = 		NULL;
	Preconditions.condition[2].parameter3 = 	NULL;
	Preconditions.condition[2].value3 = 		NULL;

	Preconditions.condition[3].condition =		PRECON_GETKEY;
	Preconditions.condition[3].parameter1 = 	PARAM_OBJECTKEY;
	Preconditions.condition[3].value1 = 		NULL;
	Preconditions.condition[3].parameter2 = 	PARAM_KEYSTATUS;
	Preconditions.condition[3].value2 = 		NULL;
	Preconditions.condition[3].parameter3 = 	NULL;
	Preconditions.condition[3].value3 = 		NULL;
	
	Preconditions.numpreconditions =			4;
	
	Effects.effect[0].condition =				EFFECT_MOVETO;
	Effects.effect[0].parameter1 = 				PARAM_OBJECTAGENT;
	Effects.effect[0].value1 = 					NULL;
	Effects.effect[0].parameter2 =	 			PARAM_XPOS;
	Effects.effect[0].value2 = 					NULL;
	Effects.effect[0].parameter3 = 				PARAM_YPOS;
	Effects.effect[0].value3 = 					NULL;
	
	Effects.effect[1].condition =				EFFECT_UNLOCK;
	Effects.effect[1].parameter1 = 				PARAM_OBJECTDOOR;
	Effects.effect[1].value1 = 					NULL;
	Effects.effect[1].parameter2 = 				PARAM_LOCKSTATUS;
	Effects.effect[1].value2 = 					NULL;
	Effects.effect[1].parameter3 =	 			NULL;
	Effects.effect[1].value3 = 					NULL;

	Effects.effect[2].condition =				EFFECT_OPEN;
	Effects.effect[2].parameter1 = 				PARAM_OBJECTDOOR;
	Effects.effect[2].value1 = 					NULL;
	Effects.effect[2].parameter2 = 				PARAM_OPENSTATUS;
	Effects.effect[2].value2 = 					NULL;
	Effects.effect[2].parameter3 =			 	NULL;
	Effects.effect[2].value3 = 					NULL;

	Effects.effect[3].condition =				EFFECT_PICKUP;
	Effects.effect[3].parameter1 = 				PARAM_OBJECTKEY;
	Effects.effect[3].value1 = 					NULL;
	Effects.effect[3].parameter2 = 				PARAM_KEYSTATUS;
	Effects.effect[3].value2 = 					NULL;
	Effects.effect[3].parameter3 = 				NULL;
	Effects.effect[3].value3 = 					NULL;
	
	Effects.numeffects =						4;
	
	
	
	}



//	--------------------------------------------------------------------------
// 	Initialise the Message Sprite 
//	--------------------------------------------------------------------------
//	Description:	Initialise a sprite for use in displaying messages
//
//	Parameters:		None
//	Returns:		None

void MessageSpriteInit()
	{
	// Local variables to use
	RECT		rect;
	GsIMAGE		tim_data;
	
	// Get TIM data
	GsGetTimInfo((u_long *)(MESSAGESPRITEADDR + 4),&tim_data);
	
	// Load the image into the frame buffer
	rect.x = tim_data.px;
	rect.y = tim_data.py;
	rect.w = tim_data.pw;
	rect.h = tim_data.ph;
	LoadImage(&rect,tim_data.pixel);
	
	// Load the CLUT info
	rect.x = tim_data.cx;
	rect.y = tim_data.cy;
	rect.w = tim_data.cw;
	rect.h = tim_data.ch;
	LoadImage(&rect,tim_data.clut);
	
	// Set sprite attributes
	MessageSprite.attribute = 0x01000000;
	MessageSprite.x			= 100;
	MessageSprite.y			= 100;
	MessageSprite.w			= 16;  //(tim_data.pw * 2);
	MessageSprite.h			= 16;  //Ctim_data.ph;
	MessageSprite.tpage		= GetTPage(1,0,tim_data.px,tim_data.py);
	MessageSprite.u			= 16;
	MessageSprite.v			= 0;
	MessageSprite.cx		= tim_data.cx;
	MessageSprite.cy		= tim_data.cy;
	MessageSprite.r			= 75;
	MessageSprite.g			= 100;
	MessageSprite.b			= 100;
	MessageSprite.mx		= 0;
	MessageSprite.my		= 0;
	MessageSprite.scalex	= ONE;
	MessageSprite.scaley	= ONE;
	MessageSprite.rotate	= 0;
	
	// Wait for LoadImage to finish
	DrawSync(0);
	}

//	--------------------------------------------------------------------------
// 	Initialise the Graphic Display
//	--------------------------------------------------------------------------
//	Description:	Initialise the Playstation graphics system
//
//	Parameters:		None
//	Returns:		None

void DisplayInit()
	{
   	// Index counter for loop
	int index;

	// Initialise the Playstation drawing system
	SetVideoMode(MODE_PAL);
	GsInitGraph(SCREEN_WIDTH,SCREEN_HEIGHT,4,0,0);
	GsDefDispBuff(0,0,0,SCREEN_HEIGHT);

	// Initialise the ordering table (OT)
	for (index=0;index<2;index++)
		{
		WorldOT[index].length = OT_LENGTH;
		WorldOT[index].org = OTTags[index];
		}
	}

//	--------------------------------------------------------------------------
// 	Initialise the Graphic System for New Drawing Cycle
//	--------------------------------------------------------------------------
//	Description:	Initialise the Playstation double buffering Ordering Table
//					system for a new drawing cycle
//
//	Parameters:		None
//	Returns:		None

void DisplayInitDrawingCycle()
	{
	// Get the current active buffer
	activeBuff = GsGetActiveBuff();

	// Set the Packet area
	GsSetWorkBase((PACKET *)GpuPacketArea[activeBuff]);

	// Clear the ordering table
	GsClearOt(0,0,&WorldOT[activeBuff]);
	}

// ----------------------------------------------------------------------
// Initialise all of the Objects in the Environment 
// ----------------------------------------------------------------------
// Description:  Set the initial values for the objects in the environment
// Parameters:   None
// Returns:      None

void ObjectInit()
	{
	

	EnvSpeed 			= SPEED;
	EnvComplexity 		= COMPLEXITY;
	EnvStartLocation 	= STARTLOCATION;

	switch (EnvStartLocation)
		{
		case 1:
			Start.x = 0;
			Start.y = 1;
			break;
		case 2:
			Start.x = 0;
			Start.y = 5;
			break;
		case 3:
			Start.x = 0;
			Start.y = 8;
			break;
		}
	
	Agent.x	= Start.x;
	Agent.y	= Start.y;
	Agent.gotkey = false;
		
	Exit.x = 9;
	Exit.y = 4;

	// Set objects depending on the complexity scenario
	switch (EnvComplexity)
		{
		case COMPLEXITY_LOW:
			Key.x = 5;
			Key.y = 7;
			
			BRIDGENUM 	= 5;
			DOORNUM 	= 0;
	
			Bridge[0].x = 3;
			Bridge[0].y = 0;
			Bridge[0].status = BRIDGERAISED;
			Bridge[0].counter = 75;
			Bridge[0].duration = (150*EnvSpeed);

			Bridge[1].x = 5;
			Bridge[1].y = 4;
			Bridge[1].status = BRIDGELOWERED;
			Bridge[1].counter = 20;
			Bridge[1].duration = (180*EnvSpeed);
	
			Bridge[2].x = 2;
			Bridge[2].y = 7;
			Bridge[2].status = BRIDGERAISED;
			Bridge[2].counter = 205;
			Bridge[2].duration = (210*EnvSpeed);
	
			Bridge[3].x = 7;
			Bridge[3].y = 9;
			Bridge[3].status = BRIDGELOWERED;
			Bridge[3].counter = 25;
			Bridge[3].duration = (195*EnvSpeed);
	
			Bridge[4].x = 8;
			Bridge[4].y = 3;
			Bridge[4].status = BRIDGELOWERED;
			Bridge[4].counter = 125;
			Bridge[4].duration = (165*EnvSpeed);
		
			break;
			
		case COMPLEXITY_MEDIUM:
			Key.x = 3;
			Key.y = 3;

			BRIDGENUM 	= 6;
			DOORNUM 	= 4;

			Door[0].x = 1;
			Door[0].y = 0;
			Door[0].doorstatus = DOORCLOSED;
			Door[0].lockstatus = DOORUNLOCKED;
	
			Door[1].x = 2;
			Door[1].y = 8;
			Door[1].doorstatus = DOORCLOSED;
			Door[1].lockstatus = DOORUNLOCKED;

			Door[2].x = 4;
			Door[2].y = 8;
			Door[2].doorstatus = DOORCLOSED;
			Door[2].lockstatus = DOORLOCKED;
	
			Door[3].x = 5;
			Door[3].y = 0;
			Door[3].doorstatus = DOORCLOSED;
			Door[3].lockstatus = DOORLOCKED;

			Bridge[0].x = 3;
			Bridge[0].y = 1;
			Bridge[0].status = BRIDGERAISED;
			Bridge[0].counter = 75;
			Bridge[0].duration = (185*EnvSpeed);

			Bridge[1].x = 1;
			Bridge[1].y = 3;
			Bridge[1].status = BRIDGELOWERED;
			Bridge[1].counter = 20;
			Bridge[1].duration = (205*EnvSpeed);
	
			Bridge[2].x = 5;
			Bridge[2].y = 6;
			Bridge[2].status = BRIDGERAISED;
			Bridge[2].counter = 205;
			Bridge[2].duration = (210*EnvSpeed);
	
			Bridge[3].x = 6;
			Bridge[3].y = 8;
			Bridge[3].status = BRIDGELOWERED;
			Bridge[3].counter = 125;
			Bridge[3].duration = (165*EnvSpeed);
	
			Bridge[4].x = 8;
			Bridge[4].y = 2;
			Bridge[4].status = BRIDGERAISED;
			Bridge[4].counter = 150;
			Bridge[4].duration = (225*EnvSpeed);
	
			Bridge[5].x = 1;
			Bridge[5].y = 5;
			Bridge[5].status = BRIDGELOWERED;
			Bridge[5].counter = 140;
			Bridge[5].duration = (175*EnvSpeed);
	
			break;
			
		case COMPLEXITY_HIGH:
			Key.x = 9;
			Key.y = 1;

			BRIDGENUM 	= 9;
			DOORNUM 	= 8;

			Door[0].x = 1;
			Door[0].y = 6;
			Door[0].doorstatus = DOORCLOSED;
			Door[0].lockstatus = DOORUNLOCKED;
	
			Door[1].x = 2;
			Door[1].y = 0;
			Door[1].doorstatus = DOORCLOSED;
			Door[1].lockstatus = DOORUNLOCKED;

			Door[2].x = 3;
			Door[2].y = 5;
			Door[2].doorstatus = DOORCLOSED;
			Door[2].lockstatus = DOORLOCKED;
	
			Door[3].x = 5;
			Door[3].y = 7;
			Door[3].doorstatus = DOORCLOSED;
			Door[3].lockstatus = DOORLOCKED;

			Door[4].x = 5;
			Door[4].y = 9;
			Door[4].doorstatus = DOORCLOSED;
			Door[4].lockstatus = DOORUNLOCKED;
	
			Door[5].x = 6;
			Door[5].y = 2;
			Door[5].doorstatus = DOORCLOSED;
			Door[5].lockstatus = DOORUNLOCKED;

			Door[6].x = 8;
			Door[6].y = 5;
			Door[6].doorstatus = DOORCLOSED;
			Door[6].lockstatus = DOORLOCKED;
	
			Door[7].x = 8;
			Door[7].y = 8;
			Door[7].doorstatus = DOORCLOSED;
			Door[7].lockstatus = DOORLOCKED;

			Bridge[0].x = 3;
			Bridge[0].y = 3;
			Bridge[0].status = BRIDGERAISED;
			Bridge[0].counter = 75;
			Bridge[0].duration = (150*EnvSpeed);

			Bridge[1].x = 2;
			Bridge[1].y = 9;
			Bridge[1].status = BRIDGELOWERED;
			Bridge[1].counter = 20;
			Bridge[1].duration = (180*EnvSpeed);
	
			Bridge[2].x = 3;
			Bridge[2].y = 7;
			Bridge[2].status = BRIDGERAISED;
			Bridge[2].counter = 205;
			Bridge[2].duration = (210*EnvSpeed);
	
			Bridge[3].x = 4;
			Bridge[3].y = 0;
			Bridge[3].status = BRIDGELOWERED;
			Bridge[3].counter = 125;
			Bridge[3].duration = (165*EnvSpeed);
	
			Bridge[4].x = 4;
			Bridge[4].y = 2;
			Bridge[4].status = BRIDGERAISED;
			Bridge[4].counter = 150;
			Bridge[4].duration = (225*EnvSpeed);
	
			Bridge[5].x = 4;
			Bridge[5].y = 6;
			Bridge[5].status = BRIDGELOWERED;
			Bridge[5].counter = 90;
			Bridge[5].duration = (195*EnvSpeed);
	
			Bridge[6].x = 6;
			Bridge[6].y = 5;
			Bridge[6].status = BRIDGERAISED;
			Bridge[6].counter = 50;
			Bridge[6].duration = (175*EnvSpeed);
	
			Bridge[7].x = 8;
			Bridge[7].y = 0;
			Bridge[7].status = BRIDGELOWERED;
			Bridge[7].counter = 10;
//			Bridge[7].duration = (275*EnvSpeed);
			Bridge[7].duration = (155*EnvSpeed);
	
			Bridge[8].x = 8;
			Bridge[8].y = 2;
			Bridge[8].status = BRIDGELOWERED;
			Bridge[8].counter = 70;
			Bridge[8].duration = (125*EnvSpeed);
	
			break;
		
		}

	Route.numsteps = 0;
	}

	